#!/usr/bin/perl - ########################### vim: set ts=4 sw=4 ai syntax=on #
###############################################################################
# IcqServ, The ICQ - IRC Gateway.                                             #
# Copyright (C) 2001 Stanislav Lechev.                                        #
# E-Mail: <firedust@vega.bg>                                                  #
#                                                                             #
# This program is free software; you can redistribute it and/or               #
# modify it under the terms of the GNU General Public License                 #
# as published by the Free Software Foundation; either version 2              #
# of the License, or (at your option) any later version.                      #
#                                                                             #
# This program is distributed in the hope that it will be useful,             #
# but WITHOUT ANY WARRANTY; without even the implied warranty of              #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               #
# GNU General Public License for more details.                                #
#                                                                             #
# You should have received a copy of the GNU General Public License           #
# along with this program; if not, write to the Free Software                 #
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. #
###############################################################################
# YOU DO NOT HAVE WORK HERE !!!                                               #
# UNLESS YOU KNOW WHAT ARE YOU DOING.                                         #
# LOOK AT icqserv.conf INSTEAD.                                               #
###############################################################################

use lib './';
use strict;
use Net::IRC;
use Net::ICQ;
use Data::Dumper;
use Socket;

my $VERSION = "0.07";
my $maxfails = 20;

	print "IcqServ version $VERSION, Copyright (C) 2001 Stanislav Lechev.\n";
	print "\t<firedust\@vega.bg>\n\n";
	print "This program was made with the great help and many ideas of\n";
	print "\tVarbin Valchev\t<dunlop\@vega.bg>\n\n";
	print "IcqServ comes with ABSOLUTELY NO WARRANTY; for details\n";
	print "see COPYING. This is free software, and you are welcome\n";
	print "to redistribute it under certain conditions;\n\n";
	printf "Perl: %.3f\nOS: %s\n", $], $^O;
	printf "Started on : %s\n", scalar localtime($^T);

	if ($ARGV[0] =~ /\-\S?h/) { &help(); }

	my $quit = 0;
	my $DEBUG = 0;
	my $boottime = 1;
	my $ConfFileName;
	my $PIDFILE;
	my (%IRCSERV, %ICQSERV, %Admin, %CONTACTS);
	my (%WHOIS, %IRCONLINE, %ICQONLINE, %SEENHASH);
	my ($irc, $icq, $conn, @AUTOJOIN, @LOGCHAN, $TMPPID);
	my ($WorkDir, $ContactFile, $LogFile, $kidpid, $servconn);
	my $icqfail=0;
	my $irccomm=0;
	my $icqcomm=0;
	my $icq_ready=0;

	if ($#ARGV < 0) {
		printf "Using default (./icqserv.conf) config file\n";
		$ConfFileName = "./icqserv.conf";
	} else {
		$ConfFileName = $ARGV[0];
	}

	$WHOIS{'can'} = 1;
	&parseconfig($ConfFileName);
	&parsecontact;
	$IRCSERV{'CONNECTED'} = 0;

#	$DEBUG=10;

	open (PIDFILE, "$WorkDir/$PIDFILE");
	$TMPPID = <PIDFILE>;
	close (PIDFILE);
	$TMPPID =~ s/(\n|\r| )//g;

	if ($TMPPID) {
		print "There is another ICQServ Started on $TMPPID.\n";
		print "If you think this is not true please remove $WorkDir/$PIDFILE\n";
		exit 0;
	}

	if ($< == 0) {
		printf "[ UID: %-5d | GID: %-5d ] Droping root privileges...\n", $<, $(;
		$< = 99;
		$( = 99;
		printf "[ UID: %-5d | GID: %-5d ] That's better :)\n", $<, $(;
	}
	
	$0 = "IcqServ v$VERSION: IRC:$IRCSERV{'NICK'} ICQ:$ICQSERV{'UIN'}";

	if (!$DEBUG && $boottime) { 
		print "Trying to fork in background.\n";
		die "can't fork: $!" unless defined($kidpid = fork());
		*STDOUT = *DUMMYHDL;
		open (PIDFILE, ">$WorkDir/$PIDFILE");
		print PIDFILE "$kidpid\n";
		close (PIDFILE);
	}
	$boottime = 0;

	if (!$kidpid || $DEBUG) {
		&tolog("ICQSERV", "========================================");
		&tolog("ICQSERV", "IcqServ started.");

# --- IRC Module -------------------------------------------------------------	
		&tolog("IRC",	"Starting Module.");
		&tolog("IRC",	"Connecting to [$IRCSERV{'SERVER'}".
				":$IRCSERV{'PORT'}]");
		&tolog("IRC",	"with nick [$IRCSERV{'NICK'}!".
				"$IRCSERV{'USER'}] and ircname ".
				"[$IRCSERV{'IRCNAME'}] ".
				"from [$IRCSERV{'ADDR'}].");
		$irc = new Net::IRC;
		$conn = $irc->newconn(	Server   => $IRCSERV{'SERVER'},
					Port     => $IRCSERV{'PORT'},
					LocalAddr=> $IRCSERV{'ADDR'}) ||
					die "can't connect to irc server ! ($!)";

		$conn->sl("PASS $IRCSERV{'SERVPASS'} :TS");
		$conn->sl("CAPAB TS3 SSJOIN");
		$conn->sl("SERVER $IRCSERV{'SERVNAME'} 1 :ICQServ v$VERSION");
		$conn->sl("SVINFO 3 1 0 :".time);
		$conn->sl("NICK $IRCSERV{'NICK'} 1 ".time." +ioO $IRCSERV{'USER'} $IRCSERV{'HOST'} $IRCSERV{'SERVNAME'} 0 :$IRCSERV{'IRCNAME'}");
		&on_connect($conn);
		$conn->{_debug} = 1 if $DEBUG > 5;

		&tolog("IRC", "Module Started.");
# --- ICQ Module -------------------------------------------------------------	
		&tolog("ICQ",	"Starting Module.");
		&icq_connect();
		$icq->{_debug} = 1 if $DEBUG > 5;
		&tolog("ICQ", "Module Started.");
# --- Handlers ---------------------------------------------------------------	
		&addhandlers();
# --- Main Loop --------------------------------------------------------------	
		while (!$quit) {
			$irc->do_one_loop();
			$icq->do_one_loop();
		}
# --- Not Reached :) ---------------------------------------------------------	
		&exitserv("0", "hmm... why am i here ?");
	}
# ----------------------------------------------------------------------------

sub icq_connect {
	my @TMP1;

	$icq->disconnect() if $icq->{_connected};
	sleep 2;
	&tolog("ICQ",	"Connecting to ICQ Server with ".
			"UIN $ICQSERV{'UIN'}.");
	$icq = Net::ICQ->new(	$ICQSERV{'UIN'}, $ICQSERV{'PASS'}, 
				$ICQSERV{'SERVER'},$ICQSERV{'PORT'}) ||
				&exitserv("0", "cant connect icq server");
	
	# we need to call connect in ICQ.pm 0.16
	# also we must add out handlers before connect

	&addhandlers();
	$icq->connect();
	$icq_ready = 0;
	while (!$icq_ready) {
		$icq->do_one_loop;
	}

	sleep 2
	&tolog("ICQ", "Sending Contact List to server.");
	delete @TMP1[0 .. $#TMP1];
	foreach (keys(%CONTACTS)) {
		push @TMP1, $_;
		if ($#TMP1 == 49) {
			$icq->send_event('CMD_CONTACT_LIST', {
				'num_contacts'	=> $#TMP1,
				'uins'		=> \@TMP1});
			delete @TMP1[0 .. $#TMP1];
		}
	}
	$icq->send_event('CMD_CONTACT_LIST', {
		'num_contacts'	=> $#TMP1,
		'uins'		=> \@TMP1});

}

sub on_user_disconnect {
	my ($irc, $event) = @_;

	my ($nick) = ($event->nick);
	my ($info) = ($event->args);


#	&tolog("TEST", "User DISConnecting: nick [$nick] | reason [$info]");
	$SEENHASH{lc($nick)} = [0, time,
		$SEENHASH{lc($nick)}[2] ? $SEENHASH{lc($nick)}[2] : '',
		$SEENHASH{lc($nick)}[3] ? $SEENHASH{lc($nick)}[3] : '',
		$SEENHASH{lc($nick)}[4] ? $SEENHASH{lc($nick)}[4] : '',
		$info];
	
}

sub on_user_nickchange {
	my ($irc, $event) = @_;

	my ($from) = ($event->from);
	my ($to)   = ($event->args)[0];
	$SEENHASH{lc($to)} = [	$SEENHASH{lc($from)}[0],
				$SEENHASH{lc($from)}[1],
				$SEENHASH{lc($from)}[2],
				$SEENHASH{lc($from)}[3],
				$SEENHASH{lc($from)}[4],
				$SEENHASH{lc($from)}[5] ];
	$SEENHASH{lc($from)} = [ 9, time, $to, '', '', ''];
}

sub whois_us {
	my ($irc, $event) = @_;

	my ($from) = ($event->from);
	my ($to)   = ($event->to);
	my ($args) = ($event->args);

	&tolog("IRC", "WHOIS: $from -> $to [$args]");
}
sub on_user_connect {
	my ($irc, $event) = @_;

	my ($host) = ($event->host);
	my ($serv) = ($event->to);
	my ($time) = ($event->format);
	my ($nick) = ($event->nick);
	my ($user) = ($event->user);
	my ($info) = ($event->args);
	$SEENHASH{lc($nick)} = [1, $time, $user, $host, $serv, $info];


#	&tolog("TEST", "User Connecting: from [$host] | to [$serv] | at [".
#			localtime($time)." ($time)] | nick [$nick] | user [".
#			"$user] | ircname [$info]");

}

sub icq_status {

	my ($stat) = @_;

	return "online" if ($stat eq 0x00);
	return "invisible" if ($stat eq 0x100);
	return "not available" if ($stat eq 0x04);
	return "not available" if ($stat eq 0x05);
	return "free for chat" if ($stat eq 0x20);
	return "occupied" if ($stat eq 0x10);
	return "occupied" if ($stat eq 0x11);
	return "away" if ($stat eq 0x01);
	return "do not disturb" if ($stat eq 0x13);
	return "do not disturb" if ($stat eq 0x02);
	return "unknown ($stat)";
}

sub icq_status_online {
	my ($icq, $event) = @_;

	my ($uin, $ip, $port, $real_ip, $status);
	my ($a,$b,$c,$d,$aa,$nickname,$host, $tmpip);
	$uin = $event->{params}->{uin};
	$aa = pack("N", $event->{params}->{ip});
	$aa = inet_ntoa($aa);
	($a,$b,$c,$d) = split(/\./,$aa,4);
	$ip = join(".",$d,$c,$b,$a);
	$port = $event->{params}->{port};
	$aa = inet_ntoa(pack("N", $event->{params}->{real_ip}));
	($a,$b,$c,$d) = split(/\./,$aa,4);
	$real_ip = join(".",$d,$c,$b,$a);
	$status = &icq_status($event->{params}->{status});
$DEBUG=5;
	&tolog("ICQ","$CONTACTS{$uin}($uin) is online with status [$status], ip: [$ip], port:[$port], realip: [$real_ip]") if ($DEBUG > 2);

	$nickname = $CONTACTS{$uin} ? $CONTACTS{$uin} : $uin;
	$host = gethostbyaddr(inet_aton($ip), AF_INET);
#	$host = "$ip.$IRCSERV{'HOST'}" if ($host=~/^$/);
	$host = "$nickname.with.unknown.host" if ($host=~/^$/);
	if (!$IRCONLINE{lc($nickname)}) {
		$conn->sl("NICK ICQ-$nickname 1 ".time." +w icquser $host $IRCSERV{'SERVNAME'} 0 :$nickname on icq");
		$conn->sl(":ICQ-$nickname JOIN #ICQ-USERS");
		$conn->sl(":ICQ-$nickname NOTICE #ICQ-USERS :is $status.");
		$IRCONLINE{lc($nickname)}=1;
		$ICQONLINE{lc($nickname)}=1;
	} else {
		&tolog("ICQ","[ICQ-$nickname] already in irc");
	}

}


sub icq_status_offline {
	my ($icq, $event) = @_;

	my ($uin,$nickname);
	
	$uin = $event->{params}->{uin};

	&tolog("ICQ","$CONTACTS{$uin}($uin): Goes Offline") if ($DEBUG > 3);
	$nickname = $CONTACTS{$uin} ? $CONTACTS{$uin} : $uin;
	&tolog("IRC", "hmmm i think there is no ICQ-$nickname nick.") if (!$IRCONLINE{lc($nickname)});
	if ($IRCONLINE{lc($nickname)}) {
		$conn->sl(":ICQ-$nickname QUIT :logged off.");
		$IRCONLINE{lc($nickname)}=0;
		$ICQONLINE{lc($nickname)}=0;
	}
}

sub icq_status_update {
	my ($icq, $event) = @_;

	my ($uin, $status,$nickname);
	
	$uin = $event->{params}->{uin};
	$status = &icq_status($event->{params}->{status});
	return 0 if ($status =~ /^unknown/);

	&tolog("ICQ","$CONTACTS{$uin}($uin): Changed his status to [$status]") if ($DEBUG > 3);
	$nickname = $CONTACTS{$uin} ? $CONTACTS{$uin} : $uin;
	if ($IRCONLINE{lc($nickname)}) {
		$conn->sl(":ICQ-$nickname PRIVMSG #ICQ-USERS :changing status to [$status]");
	} else {
		&tolog("IRC", "hmm i havent ICQ-$nickname online... trying to sign on.");
		$conn->sl("NICK ICQ-$nickname 1 ".time." +w icquser $nickname.with.unknown.host $IRCSERV{'SERVNAME'} 0 :$nickname on icq");
		$conn->sl(":ICQ-$nickname JOIN #ICQ-USERS");
		$conn->sl(":ICQ-$nickname NOTICE #ICQ-USERS :is online with status [$status].");
		$IRCONLINE{lc($nickname)}=1;
		$ICQONLINE{lc($nickname)}=0;
	}

		
}

sub icq_on_msg {
	my ($icq, $event) = @_;

	my ($in_params, $uin, $mess, $nick, $message, $nickname, $text);
	
	$in_params = $event->{params};
	$uin = $in_params->{uin};
	$message = $in_params->{text};

	if ($event->{type} == 19) {
		my (%cont) = ($event->{contacts});
		foreach (keys(%cont)) {
			&tolog("TEST","$_: $cont{'$_'}");
		}
	}
	$message =~ s/\n/ /g;
	$message =~ s/\r/ /g;

	$icqcomm++;

	&tolog("ICQ", "Received message [$message] from [$CONTACTS{$uin}($uin)].");

	$icq->send_event('CMD_SEARCH_UIN', {'uin' => $uin}) if (!$CONTACTS{$uin});

	($nick, $mess) = &parse_icq($message);

	if ($nick =~/\D/) {
		if ($nick =~ /^$IRCSERV{'NICK'}$/i) {
			&tolog("ICQ", "[$CONTACTS{$uin}($uin)] is a shitmaker!");
			$icq->send_event('CMD_SEND_MESSAGE', {
				'type' => 1,
				'text' => "Hey... Do you think i'm stupid ?!? Dont fuck with me !!!",
				'receiver_uin' => $uin});
			return 0;
		}


		&tolog("ICQ", "Message from [$CONTACTS{$uin}/$uin] for [$nick]: $mess.");
		$nickname = $CONTACTS{$uin} ? $CONTACTS{$uin} : $uin;
		if ($IRCONLINE{lc($nickname)}) {
			$conn->sl(":ICQ-$nickname PRIVMSG $nick :$mess");
		} else {
			&irc_privmsg("$nick", "$CONTACTS{$uin}($uin): $mess");
		}
		return 0;
	} elsif ($nick =~/^1$/) {
		if ($WHOIS{'can'}) {
			$mess=~s/[\n\r]//g;
			($mess) = ($mess=~/^(\S*)/m);
			&tolog("ICQ", "WHOIS for [$mess] from [$CONTACTS{$uin}($uin)]");
			$WHOIS{'can'} = 0;
			$WHOIS{'uin'} = $uin;
			$WHOIS{'nick'} = $mess;
			$WHOIS{'time'} = time;
			$conn->sl(":$IRCSERV{'NICK'} WHOIS $mess");
		} else {
			$icq->send_event('CMD_SEND_MESSAGE', {
				'type' => 1,
				'text' => "Sorry... cant whois now. Try again after few moments.",
				'receiver_uin' => $uin});
			$WHOIS{'can'} = 1 if ($WHOIS{'time'} + 10 < time);
		}
		return 0;
	} elsif ($nick =~/^2$/) {
		if (defined $SEENHASH{lc($mess)} ) {
			if ( $SEENHASH{lc($mess)}[0] == "1" ) {
				$text = "[$mess] is online";
			} elsif ( $SEENHASH{lc($mess)}[0] == "9" ) {
				$text = "[$mess] is [".$SEENHASH{lc($mess)}[2]."]";
			} else {
				$text = "[$mess!".$SEENHASH{lc($mess)}[2]."\@".$SEENHASH{lc($mess)}[3]."] was on ".$SEENHASH{lc($mess)}[4]."...\r\n";
				$text.= "quit at ".localtime($SEENHASH{lc($mess)}[1])." with reason ".$SEENHASH{lc($mess)}[5]."...";
			}
		} else {
			$text = "hmm... i can't remember [$mess]";
		}
		$icq->send_event('CMD_SEND_MESSAGE', {
			'type' => 1,
			'text' => "$text",
			'receiver_uin' => $uin});

		return 0;
	} elsif ($nick =~/^3$/) {
		$text = "Hello, $CONTACTS{$uin}.\r\n";
		$text.= "I'm ICQ-IRC Gateway made by Stanislav Lechev (AngelFire) <firedust\@vega.bg>\r\n";
		$text.= "and the great help of Varbin Valchev (dunlop) <dunlop\@vega.bg>\r\n";
		$text.= "My admin is $Admin{'NAME'} ($Admin{'NICK'}). E-Mail: <$Admin{'MAIL'}>. UIN: <$Admin{'UIN'}>.\r\n";
		$icq->send_event('CMD_SEND_MESSAGE', {
			'type' => 1,
			'text' => "$text",
			'receiver_uin' => $uin});
		return 0;

	}
	&icqhelp($uin);
}

sub on_whois {
	my ($self, $event) = @_;
	my (@args) = ($event->args);
	my ($COMM) = ($event->type);
	shift(@args);
	my ($nick1) = shift(@args);
	my ($arg) = join(' ', @args);
	my ($tmp,$tmp1,$tmp2);

#	&tolog("IRC","WHOIS($COMM): [$arg]");
	return 0 if ($WHOIS{'can'});

	if ($COMM =~ /whoisuser/) {
#		&tolog("TEST", "$arg");
		($tmp,$tmp1,$tmp2) = ($arg =~ /^(\S*) (\S*) \* (.*)/);
		$WHOIS{'res'} .= "User: $tmp\r\nHost: $tmp1\r\nIrcName: $tmp2\r\n";
	} elsif ($COMM =~ /whoisoperator/) {
		$WHOIS{'res'} .= "Oper: $arg\r\n";
	} elsif ($COMM =~ /whoischannels/) {
		$WHOIS{'res'} .= "Channels: $arg\r\n";
	} elsif ($COMM =~ /whoisidle/) {
		($tmp,$tmp1) = ($arg=~/^(\d*) (\d*)/);
		$WHOIS{'res'} .= "Idle: $tmp sec\r\nSignOn: ".scalar localtime($tmp1)." ($tmp1)";
	} elsif ($COMM =~ /away/) {
		$WHOIS{'res'} .= "Away: $arg\r\n";
	}		
	if ($COMM =~ /endofwhois/) {
#		&tolog("ICQSERV", "Sending to [$WHOIS{'uin'}]: [$WHOIS{'res'}]");
		$icq->send_event('CMD_SEND_MESSAGE', {
				'type' => 1,
				'text' => "Nick: $nick1\r\n$WHOIS{'res'}",
                                'receiver_uin' => $WHOIS{'uin'}
		});
		$WHOIS{'can'} = 1;
		$WHOIS{'res'} = "";
	} elsif ($COMM =~ /nosuchnick/) {
		$icq->send_event('CMD_SEND_MESSAGE', {
				'type' => 1,
				'text' => "[$nick1]: no such nick\r\n",
				'receiver_uin' => $WHOIS{'uin'}
		});
		$WHOIS{'can'} = 1;
		$WHOIS{'res'} = "";
	}
	
	return 0;
}

sub parse_icq {
	my ($message) = @_;
	my ($nick, $mess,$comm);
	
	if ($message=~/^\s*(\S*) (\S*)\s*(.*)/i) {
		$comm = "$1";
		$nick = "$2";
		$mess = "$3";
	}
	if ($comm =~ /send/i) {
		return 0 if (!$nick || !$mess);
		return($nick, $mess);
	} elsif ($comm =~ /whois/i) {
		return 0 if (!$nick);
		return ("1","$nick");
	} elsif ($comm =~ /seen/i) {
		return 0 if (!$nick);
		return ("2","$nick");
	} elsif ($comm =~ /about/i) {
		return ("3","");
	}
	return 0;
}

sub icq_user_found {
	my ($icq, $event) = @_;
	my ($uin) = ($event->{params}->{uin});
	my ($nick) = ($event->{params}->{nickname});
	my (@needed);

	$nick=~s/ //g;
	if (!$CONTACTS{$uin}) {
		&tolog("ICQ", Dumper($event)) if $DEBUG > 5;
		&tolog("ICQ", "NewContact: [$uin] nick: [$nick].");
		$CONTACTS{$uin} = "$nick";
		open(CONTFILE, ">>$ContactFile") || warn("Error opening \"$ContactFile\" for write.");
		print CONTFILE "$uin $nick\n";
		close(CONTFILE);
		@needed = ("$uin");
		$icq->send_event('CMD_CONTACT_LIST', {
			'num_contacts'	=> 1,
			'uins'		=> \@needed});

	}

}

sub on_msg {
	my ($irc, $event) = @_;
	my ($nick) = $event->nick;
	my ($arg) = ($event->args);
	my ($to) = ($event->to)[0];
	my ($params);
	my ($text,$uin,$mess);
	my ($adcomm, $adpass, $null);
	my $stats=0;

	my (@argz) = (split(' ', $arg));

	$irccomm++;

	if (($nick =~ /^$Admin{'NICK'}$/i) && ($argz[0] =~ /^admin/i)) {
		$null = shift(@argz);
		$adcomm = shift (@argz);
		$adpass = shift (@argz);

		if (!($Admin{'PASS'} =~ /^$adpass$/)) {
			&tolog("IRC", "[$nick] admin attempt with wrong pass [$adpass].");
			&irc_notice("$nick", "Wrong Password.");
			&irc_notice("$nick", "Syntax: admin command password parameters");
			return 0;
		}
		$stats++;
		if ($adcomm =~ /^quit$/i) {
			&exitserv("0", @argz ? "@argz":"Requested by $Admin{'NICK'}.");
			return 0;
		}
		elsif ($adcomm =~ /^raw$/i) {
			&tolog("IRC", "[$nick] raw command: [@argz]");
			$conn->sl("@argz");
			return 0;
		}
		elsif ($adcomm =~ /^debug$/i) {
			$DEBUG = @argz[0];
			$conn->{_debug} = 1 if $DEBUG > 5;
			$conn->{_debug} = 1 if $DEBUG > 5;
			$icq->{_debug} = 0 if $DEBUG < 6;
			$icq->{_debug} = 0 if $DEBUG < 6;
			&tolog("IRC", "[$nick] changed debug level to: [".@argz[0]."].");
			&irc_notice("$nick", "Changed Debug Level: ".@argz[0]."");
			return 0;
		}
		elsif ($adcomm =~ /^stats$/i) {
			&tolog("IRC", "[$nick] requestet admin statistics");
			&irc_notice("$nick", "Current statisrics:");
			&irc_notice("$nick", "     Debug Level    : $DEBUG");
			&irc_notice("$nick", "     Admin Commands : $stats");
			&irc_notice("$nick", "Connection Failures : $icqfail");
			$null = keys (%CONTACTS);
			&irc_notice("$nick", "     Contact List   : $null uins");
			$null=0;
			foreach (keys(%IRCONLINE)) {
				$null++ if $IRCONLINE{$_};
			}
			&irc_notice("$nick", "     IRC Users      : $null");
			$null=0;
			foreach (keys(%ICQONLINE)) {
				$null++ if $ICQONLINE{$_};
			}
			&irc_notice("$nick", "     ICQ Users      : $null");

		}
		elsif ($adcomm =~ /^rehash$/i) {
			&tolog("ICQSERV", "[$nick] is rehashing me.");
			&parseconfig($ConfFileName);
			&parsecontact;
		}
		else {
			&tolog("ICQSERV", "[$nick] unknown command $adcomm");
			&irc_notice("$nick", "Syntax: admin command password parameters");
			&irc_notice("$nick", "Commands: quit, rehash, raw, stats, debug.");
			return 0;
		}

		return 0;
	}


	&tolog("IRC", "Message from [$nick] to [$to]: $arg");

	if ($to =~ /^$IRCSERV{'NICK'}$/i) {
		&tolog("IRC", "Message to [$IRCSERV{'NICK'}]");
		if ( $arg =~ /^seen\s*(\S*)/i ) {
			if (defined $SEENHASH{lc($1)} ) {
				if ( $SEENHASH{lc($1)}[0] == "1" ) {
					&irc_notice("$nick", "$1 is online");
				} elsif ( $SEENHASH{lc($1)}[0] == "9" ) {
					&irc_notice("$nick", "$1 is ".
							$SEENHASH{lc($1)}[2]."");
				} else {
					&irc_notice("$nick", "$1!".$SEENHASH{lc($1)}[2]."\@".$SEENHASH{lc($1)}[3]." was on ".$SEENHASH{lc($1)}[4]."...");
					&irc_notice("$nick", "quit at ".localtime($SEENHASH{lc($1)}[1])." with reason ".$SEENHASH{lc($1)}[5]."...");
				}
			} else {
				&irc_notice("$nick", "hmm... i can't remember $1");
			}
		return 0;
		}
		($uin,$mess) = &parse_irc($arg);
	}
	elsif ($to =~ /^ICQ-/i) {
		($uin) = ($to =~ /^ICQ-(\S*)$/i);
		foreach (keys %CONTACTS) {
			$uin = $_ if ($CONTACTS{$_} =~ /^$uin$/i);
			$mess = $arg;
		}
		&irc_notice("$nick", "Report this error to $Admin{'NICK'} please.") if (!$uin || !$mess);
	} else {
		&irc_notice("$nick", "Hey... are you talking to me?");
	}


	if ($uin =~/^\d{6,}$/) {
		if ($uin =~ /^$ICQSERV{'UIN'}$/i) {
			&irc_notice("$nick", "Hey ... dont be stupid man!");
			return 0;
		}
		&tolog("IRC", "Message from [$nick] for [$CONTACTS{$uin}($uin)]: $mess");
		$text = "Message from [$nick] at $IRCSERV{'NETNAME'}\r\n".
			"---\r\n$mess\r\n---\r\n".
			"Reply with: \"send $nick <msg>\" or \"help\" for help.";
		$params = {
			'type'		=> 1,
			'text'		=> "$text",
			'receiver_uin' 	=> "$uin"
		};
		$icq->send_event('CMD_SEND_MESSAGE', $params);
		return 0;
	} elsif ($uin =~/^\d{2,}$/) {
		&irc_notice("$nick", "Sorry UIN must be min. 6 digits.");
		return 0;
	} else {
#		&tolog("IRC", "HELP: [$nick]: [$arg]");
		&irchelp($uin, $nick, $mess);
		return 0;
	}

}

sub irchelp {
	my ($what, $nick, $mess) = @_;
	my ($text,$tmp,$c,$isc,@keys);
	my $stats=0;

	if ($what =~ /1/) {
		&tolog("IRC","[$nick]: No such uin/contact [$mess]");
		$text = "There is no such nick in my contact list. Try ".
			"\"/msg $IRCSERV{'NICK'} show [nick]\" to ".
			"show contacts starting with \"nick\".";
		&irc_notice("$nick", "$text");
	} elsif ($what =~ /2/) {
		&tolog("IRC","[$nick] wanted to see Contact List for [$mess]");
		$c=0;
		$isc=0;
		@keys = sort {
			$CONTACTS{$a} cmp $CONTACTS{$b}
		} keys %CONTACTS;

		foreach $tmp(@keys) {
			if (($CONTACTS{$tmp} =~ /^$mess/i) && ($c <10)) {
				$isc=1;
				$c++;
				&irc_notice("$nick", "$CONTACTS{$tmp} = $tmp");
				sleep 1;
			}
			if ($c == 10) {
				$c++;
				&irc_notice("$nick", "There are more than 10 entry. Be more specific.");
				sleep 1;
			}
			
		}
		if ( $isc == 0 ) {
			$c++;
			&irc_notice("$nick", "There is no entry starting with [$mess]");
		}
	} elsif ($what =~ /3/) {
		&tolog("IRC","[$nick]: about message");
		&irc_notice("$nick", "Hello, $nick.");
		&irc_notice("$nick", "I'm ICQ-IRC Gateway made by Stanislav Lechev (AngelFire) <firedust\@vega.bg>");
		sleep 1;
		&irc_notice("$nick", "and the great help of Varbin Valchev (dunlop) <dunlop\@vega.bg>");
		&irc_notice("$nick", "My admin is $Admin{'NAME'} ($Admin{'NICK'}). E-Mail: <$Admin{'MAIL'}>. UIN: <$Admin{'UIN'}>.");
	} elsif ($what =~ /4/) {
		&irc_notice("$nick", "Current statistics:");
		&irc_notice("$nick", "     Commands via IRC: $irccomm");
		&irc_notice("$nick", "     Commands via ICQ: $icqcomm");
		$stats = $icqcomm + $irccomm;
		&irc_notice("$nick", "     Commands total  : $stats");
	} else {
		&tolog("IRC","[$nick] wanted help.");
		&irc_notice("$nick", "ICQ - IRC Gateway by AngelFire");
		&irc_notice("$nick", "==============================");
		&irc_notice("$nick", "send <uin> <msg>   - send <msg> to <uin>");
		&irc_notice("$nick", "send <nick> <msg>  - send <msg> to <nick> (from contact list)");
		&irc_notice("$nick", "show <nick>        - show contacts starting with <nick>");
		&irc_notice("$nick", "about              - show info about this bot.");
		&irc_notice("$nick", "admin              - administrative commands.");
		&irc_notice("$nick", "stats              - show statistics about received commands.");
		&irc_notice("$nick", "seen <nick>        - info about nick");
		&irc_notice("$nick", "==============================");
	}
}

sub parse_irc {
	my ($message) = @_;
	my ($comm, $uin, $mess, $tmp);
	
	if ($message=~/^\s*(\S*)\s*(\S*) (.*)/i) {
		$comm = "$1";
		$uin = "$2";
		$mess = "$3";
	} elsif ($message=~/^\s*(\S*)\s*(\S*)/) {
		$comm = "$1";
		$mess = "$2";
	}

	if ($comm =~ /^send/i) {
		return 0 if (!$uin || !$mess);
		return($uin, $mess) if ($uin=~/^\d*$/);
		foreach $tmp(keys %CONTACTS) {
			return($tmp, $mess) if ($CONTACTS{$tmp} =~ /^$uin$/i);
		}
		return("1","");
	} elsif ($comm =~ /^show/i) {
		return("2","$mess");
	} elsif ($comm =~ /^about/i) {
		return("3","$mess");
	} elsif ($comm =~ /^stats/i) {
		return("4","$mess");
	}
	return "0";
}


sub icqhelp {
	my ($uin) = @_;
	my ($params,$text);

	&tolog("ICQ","[$CONTACTS{'$uin'}($uin)] wanted help.");
	$text = "Hi, I'm $IRCSERV{'NICK'} on $IRCSERV{'NETNAME'}.\r\n\r\n";
	$text.= "Commands:\r\n";
	$text.= "send <nick> <message> - sends <message> to <nick> on irc.\r\n";
	$text.= "whois <nick>          - will give you /whois <nick>\r\n";
	$text.= "seen <nick>           - will give you info about <nick>\r\n";
	$text.= "about                 - about me.\r\n";
	$text.= "\r\nExample: \"send AngelFire Hello, How are you?\".\r\n";
	$text.= "This will send \"Hello, How are you?\" to \"AngelFire\" on irc";
	$params = {
		'type'          => 1,
		'text'          => "$text",
		'receiver_uin'  => $uin
	};
	$icq->send_event('CMD_SEND_MESSAGE', $params);
}

sub wallops {
	my ($what) = @_;
	$conn->sl(":$IRCSERV{'NICK'} WALLOPS :$what");
	&tolog("IRC", "WALLOPS: $what");
}

sub icq_on_noconn {
	my $left = 0;
	
	$icqfail++;
	$left = $maxfails - $icqfail;
	&exitserv("1", "Giving up... no connection with icq server") if ($left < 0);
#	&wallops("There is no connection with icq server. Reconnecting... (Failures: $icqfail; Giving up after $left)");
	&icq_connect;
	
}

sub icq_on_goaway {
	my $left = 0;
	$icqfail++;
	$left = $maxfails - $icqfail;
	&exitserv("1", "Server forced us to disconnect.") if ($left < 0);
#	&wallops("ICQ Server forced us to disconnect. Reconnecting... (Failures: $icqfail; Giving up after $left)");
	&icq_connect;
}

sub icq_on_badpass {
	&exitserv("1", "ICQServ: Bad Password.");
}

sub on_connect {
	my $irc = shift;
	my $chan;

	&tolog("IRC", "Connected to $IRCSERV{'SERVER'}");
	&tolog("IRC", "Joining #icq-debug...");
	$irc->sl(":$IRCSERV{'SERVNAME'} SJOIN 0 #icq-debug +tnsk $IRCSERV{'SERVPASS'} :\@$IRCSERV{'NICK'}");
	$irc->sl(":$IRCSERV{'NICK'} TOPIC #icq-debug :ICQServ Debug Channel");

	&tolog("IRC", "Joining #ICQ-USERS...");
	$irc->sl(":$IRCSERV{'SERVNAME'} SJOIN 0 #ICQ-Users +tn :\@$IRCSERV{'NICK'}");
	$irc->sl(":$IRCSERV{'NICK'} TOPIC #ICQ-Users :Here you can find the users of my contact list which are online in ICQ");
	foreach $chan(@AUTOJOIN) {
		&tolog("IRC", "AutoJoining $chan...");
		$irc->sl(":$IRCSERV{'NICK'} JOIN $chan");
		&irc_privmsg("$chan", "Hi there.");
	};

	$IRCSERV{'CONNECTED'} = 1;	
}



sub on_nick_taken {
	my ($self) = shift;

	&tolog("IRC","nickname ".$self->nick." in use, killing him. ?!?!?!?");
	$self->nick($self->nick."_");
}


sub on_disconnect {
	my ($self, $event) = @_;

	$IRCSERV{'CONNECTED'} = 0;

	&tolog("IRC","connection lost with ".$event->from()." (".($event->args())[0].").");

	&exitserv("100","connection lost with ".$event->from()." (".($event->args())[0].")");
}

sub on_nick_error {
	my ($irc, $event) = @_;
	my ($line) = ($event->args());
	my ($from) = ($event->from());
	my ($to) = ($event->to);
	my ($reason)   = ($event->format);
	my ($nick);

	if ($to =~ /^ICQ-/i) {
		($nick) = ($to =~ /^ICQ-(\S*)$/i);
		if ($IRCONLINE{lc($nick)}) {
			$IRCONLINE{lc($nick)}=0;
			&tolog("IRC","[$nick] cannot login [$reason]");
		} else {
			&tolog("IRC","not for me [$from] [$to] [$reason] [$line]");
		}
	} else {
		&tolog("IRC","hmm... [$from] [$to] [$reason] [$line]");
	}
	
}

sub on_kill {
	my ($irc, $event) = @_;
	my ($line) = ($event->args());
	my ($from) = ($event->from());
	my ($to)   = ($event->to)[0];
	my ($nick);

	if ($to =~ /^ICQ-/i) {
		($nick) = ($to =~ /^ICQ-(\S*)$/i);
		if ($IRCONLINE{lc($nick)}) {
			$IRCONLINE{lc($nick)}=0;
			&tolog("IRC","[$from] killed icquser [$nick] for [$line]");
		} else {
			&tolog("IRC","[$from] killed [$to] for [$line]");
		}
	} else {
		&tolog("IRC","[$to] has been killed by [$from] for [$line]");
	}
	
}


sub on_init {
	my ($self, $event) = @_;
	my (@args) = ($event->args);
	my ($COMM) = ($event->type);
	shift (@args);
	
	foreach(@args) {
	&tolog("IRC","($COMM) - $_");
	}
}


sub on_public {
	my ($self, $event) = @_;
	my $to = join(", ", @{$event->{'to'}});
	my ($nick) = ($event->nick);
	my ($arg) = ($event->args);

	foreach (@LOGCHAN) {
#		&tolog("IRC","$to ? $_");
		&tolog("IRC","<$nick@".$to."> $arg") if ($to=~/^$_$/);
	}
}

sub on_action {
        my ($self, $event) = @_;
        my ($nick, @args) = ($event->nick, $event->args);
#        shift @args;

        &tolog("IRC"," * [$nick] @args") if $DEBUG > 2;
}

sub on_kick {
        my $self = shift;
        my $event = shift;

#        &tolog("IRC","".$event->nick." (".$event->userhost.") kicked me from ".$event->{args}[0]."!!");
}

sub on_join {
        my ($self, $event) = @_;
        my ($channel) = ($event->to)[0];
        my ($line, $host);

        $line=" * [".$event->nick."] (".$event->userhost.") joined ".$channel.".";
        &tolog("IRC", $line) if $DEBUG > 2;
}

sub on_part {
        my ($self, $event) = @_;
        my ($channel) = ($event->to)[0];
        my $line;

        $line=" * [".$event->nick."] (".$event->userhost.") left ".$channel.".";
        &tolog("IRC",$line) if $DEBUG > 2;
}

sub handle_x2 {
	my ($event) = @_;

	&tolog("ICQ", ":X2");
	$icq->send_event('CMD_ACK_MESSAGES');
}

sub icq_on_login_reply {
  my ($icq, $event) = @_;

	$icq_ready = 1;
	&tolog("ICQ", "Logged in.");

}
sub dump_event {
	my ($icq, $event) = @_;

	&tolog("ICQ", Dumper($event)) if $DEBUG > 5;
}


sub addhandlers {
	&tolog("IRC", "building all connection handlers");
	$conn->add_global_handler('376', \&on_connect);     # global
	$conn->add_global_handler('disconnect', \&on_disconnect);
	$conn->add_global_handler([ 251,252,253,254,302,255 ], \&on_init);
	$conn->add_global_handler([ 311,319,301,313,317,318,401 ], \&on_whois);
	$conn->add_global_handler(432, \&on_nick_error);
	$conn->add_global_handler(433, \&on_nick_taken);
	$conn->add_handler('caction', \&on_action);
	$conn->add_handler('public', \&on_public);
	$conn->add_handler('msg', \&on_msg);
	$conn->add_handler('kick', \&on_kick);
	$conn->add_handler('join', \&on_join);
	$conn->add_handler('part', \&on_part);
	$conn->add_handler('kill', \&on_kill);
	$conn->add_handler('quit', \&on_user_disconnect);
	$conn->add_handler('clientconnect', \&on_user_connect);
	$conn->add_handler('whois_us', \&whois_us);
	$conn->add_handler('nick', \&on_user_nickchange);
	&tolog("ICQ", "building all connection handlers");
	foreach (values(%Net::ICQ::srv_codes)) {
		$icq->add_handler($_, \&dump_event);
	}
	$icq->add_handler('SRV_LOGIN_REPLY', \&icq_on_login_reply);
	$icq->add_handler('SRV_BAD_PASS', \&icq_on_badpass);
	$icq->add_handler('SRV_GO_AWAY', \&icq_on_goaway);
	$icq->add_handler('SRV_SYSTEM_MESSAGE', \&icq_on_sysmsg);
	$icq->add_handler('SRV_SYS_DELIVERED_MESS', \&icq_on_msg);
	$icq->add_handler('SRV_USER_ONLINE', \&icq_status_online);
	$icq->add_handler('SRV_USER_OFFLINE', \&icq_status_offline);
	$icq->add_handler('SRV_STATUS_UPDATE', \&icq_status_update);
	$icq->add_handler('SRV_USER_FOUND', \&icq_user_found);
	$icq->add_handler('SRV_INFO_REPLY', \&icq_user_found);
	$icq->add_handler('SRV_NOT_CONNECTED', \&icq_on_noconn);

	$SIG{"TERM"} = sub {
		syswrite(STDOUT, "ouch. SIGTERM.\n");
		&exitserv("0", "SIGTERM Received.");
	};
	$SIG{"SEGV"} = sub {
		syswrite(STDOUT, "ouch. SIGSEGV.\n");
		&exitserv("0", "SIGSEGV Reveiced from Perl ?!?");
	};
	$SIG{"HUP"} = sub {
		syswrite(STDOUT, "ouch. SIGHUP. Rehashing.");
		&tolog("ICQSERV", "SIGHUP. Rehashing.");
		&parseconfig($ConfFileName);
		&parsecontact;
	};	
	$SIG{"INT"} = sub {
		syswrite(STDOUT, "ouch. SIGINT.\n");
		&exitserv("0", "Ctrl-C Pressed.");
	};
}

sub help {
	print "usage:\n\t$0 [config_file]\n";
	exit 0;
}

sub tolog {
	my ($type) = shift;
	my ($what) = shift;

	my ($time,$sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst);

	$what =~ s/(\n|\r)/ /g;
	$conn->sl(":$IRCSERV{'NICK'} PRIVMSG #icq-debug :$type: $what") if $IRCSERV{'CONNECTED'};
	print "[$type]: $what\n" if $DEBUG;
	($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time);
	$mon = $mon + 1;
	open(LOGFILE, ">>$WorkDir/$LogFile");
	print LOGFILE "".$mday."/".$mon." ".$hour.":".$min.":".$sec.": [$type]: $what\n";
	close(LOGFILE);
}

sub exitserv {
	my ($err) = shift;
	my ($reason) = @_;

	&tolog("IRC", "QUIT: $reason");
	$conn->sl(":$IRCSERV{'NICK'} WALLOPS :DISCONNECTING: $reason");
	$conn->sl(":$IRCSERV{'NICK'} GLOBOPS :DISCONNECTING: $reason");
	$conn->sl(":$IRCSERV{'NICK'} QUIT :$reason [irc: $irccomm][icq: $icqcomm]");

	&tolog("ICQ", "QUIT: $reason");
	$icq->disconnect();
	sleep 2;
	&tolog("ICQSERV", "========================================");
	open (PIDFILE, ">$WorkDir/$PIDFILE");
	print PIDFILE "";
	close (PIDFILE);
	unlink "$WorkDir/$PIDFILE" ||
		&tolog("EXIT", "Cant unlink pid file. Delete it manualy.");
	exit($err);
}


sub parseconfig {
	my (@CONF);
	my ($line);
	my ($tmp);

	&tolog("ICQSERV", "Parsing ConfigFile '@_'.");
	open(CONFFILE, "@_") || die("Error opening \"@_\" for reading.");
	@CONF = <CONFFILE>;
	close(CONFFILE);
	foreach $line(@CONF) {
		next if $line=~/^#/; next unless $line=~/\S/;
		if ($line =~ /^Debug/i) {
			($DEBUG) = ($line=~/^Debug\s*=\s*(\S*)/i);
			$DEBUG = 0 unless ($DEBUG=~/(1|on)/i);
		}
		elsif ($line =~ /^WorkDir/i) {
			($WorkDir) = ($line=~/^WorkDir\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^PidFile/i) {
			($PIDFILE) = ($line=~/^PidFile\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^ContactFile/i) {
			($ContactFile) = ($line=~/^ContactFile\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^LogFile/i) {
			($LogFile) = ($line=~/^LogFile\s*=\s*(.*)$/i);
		}
		elsif ($line =~ /^AdminName/i) {
			($Admin{'NAME'}) = ($line=~/^AdminName\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^AdminNick/i) {
			($Admin{'NICK'}) = ($line=~/^AdminNick\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^AdminPass/i) {
			($Admin{'PASS'}) = ($line=~/^AdminPass\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^AdminMail/i) {
			($Admin{'MAIL'}) = ($line=~/^AdminMail\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^AdminUIN/i) {
			($Admin{'UIN'}) = ($line=~/^AdminUIN\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^ServerPass/i) {
			($IRCSERV{'SERVPASS'}) = ($line=~/^ServerPass\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^BindAddress/i) {
			($IRCSERV{'ADDR'}) = ($line=~/^BindAddress\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^ServerName/i) {
			($IRCSERV{'SERVNAME'}) = ($line=~/^ServerName\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IrcNetworkName/i) {
			($IRCSERV{'NETNAME'}) = ($line=~/^IrcNetworkName\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IrcServer/i) {
			($IRCSERV{'SERVER'}) = ($line=~/^IrcServer\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IrcPort/i) {
			($IRCSERV{'PORT'}) = ($line=~/^IrcPort\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^BotNick/i) {
			($IRCSERV{'NICK'}) = ($line=~/^BotNick\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^BotUser/i) {
			($IRCSERV{'USER'}) = ($line=~/^BotUser\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^BotIrcName/i) {
			($IRCSERV{'IRCNAME'}) = ($line=~/^BotIrcName\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^BotHost/i) {
			($IRCSERV{'HOST'}) = ($line=~/^BotHost\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^WantOper\s/i) {
			($IRCSERV{'OPER'}) = ($line=~/^WantOper\s*=\s*(\S*)/i);
			$IRCSERV{'OPER'} = 0 unless ($IRCSERV{'OPER'}=~/(1|on)/i);
		}
		elsif ($line =~ /^OperNick/i) {
			($IRCSERV{'OPERNICK'}) = ($line=~/^OperNick\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^OperPass/i) {
			($IRCSERV{'OPERPASS'}) = ($line=~/^OperPass\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IcqUIN/i) {
			($ICQSERV{'UIN'}) = ($line=~/^IcqUIN\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IcqPass/i) {
			($ICQSERV{'PASS'}) = ($line=~/^IcqPass\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IcqServer/i) {
			($ICQSERV{'SERVER'}) = ($line=~/^IcqServer\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^IcqPort/i) {
			($ICQSERV{'PORT'}) = ($line=~/^IcqPort\s*=\s*(.*)/i);
		}
		elsif ($line =~ /^AutoJoin/i) {
			($tmp) = ($line=~/^AutoJoin\s*=\s*(.*)/i);
			$tmp =~ s/,/ /g;
			$tmp =~ s/\s / /g;
			@AUTOJOIN = split(" ",$tmp);
		}
		elsif ($line =~ /^LogChannels/i) {
			($tmp) = ($line=~/^LogChannels\s*=\s*(.*)/i);
			$tmp =~ s/,/ /g;
			$tmp =~ s/\s+/ /g;
			@LOGCHAN = split(" ",$tmp);
		}

	}
	&tolog("ICQSERV", "ConfigFile '@_' Done.");
}

sub parsecontact {
	my (@CONT);
	my ($line);
	my ($tmp,$c);

	&tolog("ICQSERV", "Parsing ContactFile '$ContactFile'.");
	open(CONTFILE, "$ContactFile") || warn("Error opening \"$ContactFile\" for reading.");
	@CONT = <CONTFILE>;
	close(CONTFILE);
	$c = 0;
	foreach $line(@CONT) {
		$c++;
		next if $line=~/^#/; next unless $line=~/\S/;
		if ($line =~ /^(\d*)\s*(\S*)$/i) {
			$CONTACTS{$1} = "$2";
		} else {
			&tolog("ICQSERV", "Invalid contact line ($ContactFile:$c).");
		}
	}
	
	&tolog("ICQSERV", "ContactFile '$ContactFile' Done.");
}

sub irc_notice {
	my ($to, $line) = @_;

	$conn->sl(":$IRCSERV{'NICK'} NOTICE $to :$line");
}
sub irc_privmsg {
	my ($to, $line) = @_;

	$conn->sl(":$IRCSERV{'NICK'} PRIVMSG $to :$line");
}


